# This is free and unencumbered software released into the public domain.
#
# Anyone is free to copy, modify, publish, use, compile, sell, or
# distribute this software, either in source code form or as a compiled
# binary, for any purpose, commercial or non-commercial, and by any
# means.
#
# In jurisdictions that recognize copyright laws, the author or authors
# of this software dedicate any and all copyright interest in the
# software to the public domain. We make this dedication for the benefit
# of the public at large and to the detriment of our heirs and
# successors. We intend this dedication to be an overt act of
# relinquishment in perpetuity of all present and future rights to this
# software under copyright law.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# For more information, please refer to <http://unlicense.org/>

import copy
import os
import shutil
import tempfile
from typing import Dict
from xml.etree import ElementTree as ET
import zipfile

XMLNS = {
    "wpc": "http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas",
    "cx": "http://schemas.microsoft.com/office/drawing/2014/chartex",
    "cx1": "http://schemas.microsoft.com/office/drawing/2015/9/8/chartex",
    "cx2": "http://schemas.microsoft.com/office/drawing/2015/10/21/chartex",
    "cx3": "http://schemas.microsoft.com/office/drawing/2016/5/9/chartex",
    "cx4": "http://schemas.microsoft.com/office/drawing/2016/5/10/chartex",
    "cx5": "http://schemas.microsoft.com/office/drawing/2016/5/11/chartex",
    "cx6": "http://schemas.microsoft.com/office/drawing/2016/5/12/chartex",
    "cx7": "http://schemas.microsoft.com/office/drawing/2016/5/13/chartex",
    "cx8": "http://schemas.microsoft.com/office/drawing/2016/5/14/chartex",
    "mc": "http://schemas.openxmlformats.org/markup-compatibility/2006",
    "aink": "http://schemas.microsoft.com/office/drawing/2016/ink",
    "am3d": "http://schemas.microsoft.com/office/drawing/2017/model3d",
    "o": "urn:schemas-microsoft-com:office:office",
    "r": "http://schemas.openxmlformats.org/officeDocument/2006/relationships",
    "m": "http://schemas.openxmlformats.org/officeDocument/2006/math",
    "v": "urn:schemas-microsoft-com:vml",
    "wp14": "http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing",
    "wp": "http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing",
    "w10": "urn:schemas-microsoft-com:office:word",
    "w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
    "w14": "http://schemas.microsoft.com/office/word/2010/wordml",
    "w15": "http://schemas.microsoft.com/office/word/2012/wordml",
    "w16cid": "http://schemas.microsoft.com/office/word/2016/wordml/cid",
    "w16se": "http://schemas.microsoft.com/office/word/2015/wordml/symex",
    "wpg": "http://schemas.microsoft.com/office/word/2010/wordprocessingGroup",
    "wpi": "http://schemas.microsoft.com/office/word/2010/wordprocessingInk",
    "wne": "http://schemas.microsoft.com/office/word/2006/wordml",
    "wps": "http://schemas.microsoft.com/office/word/2010/wordprocessingShape",
}
for prefix, url in XMLNS.items():
    ET.register_namespace(prefix, url)

def SubElementNS(parent: ET.Element, ns: str, name: str) -> ET.Element:
    return ET.SubElement(parent, '{' + XMLNS[ns] + '}' + name)

class MailMerge:
    """
    Context manager for executing a mail merge of a template file.
    
    Using this as a context manager means that the template file does not have
    to be decompressed or completely copied every time. Instead we hold onto
    some in-memory copies of the relevant data so we can easily construct a
    new output file.
    """

    def __init__(self, template_path: str):
        """
        Prepare a new MailMerge.
        
        The init function does nothing except storing parameters.
        Preparation of files etc. only takes place once the context manager
        is entered.
        """
        self._template_path = template_path

    def __enter__(self):
        # create a temporary directory to initially extract the zip file into
        tempdir = tempfile.mkdtemp()
        shutil.unpack_archive(self._template_path, tempdir, 'zip')

        document_path = os.path.join(tempdir, 'word', 'document.xml')

        # Parse the actual document content into memory and keep it there.
        # From this, deepcopies will be used to create the output documents.
        self._template_document = ET.parse(document_path)

        # the "Ignorable" attribute causes errors because etree only adds the
        # namespaces that are actually used and also assigns them different names.
        # So the ignorable namespaces are not reinistituted anyway.
        self._template_document.getroot().attrib.pop('{' + XMLNS['mc'] + '}Ignorable', None)

        # Now remove the document contents and re-pack the zip archive. Because
        # the word/document.xml file is absent, we can use the zipfile library to
        # later add this file into it. (Modifying a member of a zip archive is
        # not possible though.)
        os.remove(document_path)
        with tempfile.NamedTemporaryFile() as fp:
            self._template_zip = shutil.make_archive(fp.name, 'zip', tempdir)

        return self._merge

    def __exit__(self, exc, value, tb):
        # clean up temporary files and data
        os.remove(self._template_zip)
        del self._template_zip
        del self._template_document

    def _merge(self, field_values: Dict[str, str], output_path: str):
        """
        Perform the mail merge using the initial template and the given field values.
        
        If there is an error while processing the template, the resulting file will not be created.
        """
        tree = copy.deepcopy(self._template_document)

        for sdt in tree.findall('.//w:sdt', XMLNS):
            tag = sdt.find('./w:sdtPr/w:tag', XMLNS)
            # If there is no tag name, this field cannot be processed.
            if tag is None:
                continue
            tag = tag.attrib['{' + XMLNS['w'] + '}val']
            # Fields that we aren't given cannot be filled out, leave it as is.
            if tag not in field_values:
                continue

            content = sdt.find('./w:sdtContent', XMLNS)
            if content is None:
                continue

            # According to ECMA-376 Part 1 §§ 17.5.2.33 until including 17.5.2.36,
            # there are different possibilities for what the sdtContent contains:
            # - cell-level content (table cell, <w:tc>)
            # - block-level content (<w:p> or <w:tbl>)
            # - row-level content (<w:tr>)
            # - inline-level content (<w:r> and others)
            # None of these are easily enumerated tag lists, so for simplicity this
            # script only supports either a paragraph or runs of text.

            # The field can either contain a whole paragraph (e.g. if it is on a line
            # of its own). Alternatively, the sdt is inside a paragraph in which case
            # it would not be valid against the schema to add another paragraph tag.
            is_paragraph = content.find('./w:p', XMLNS) is not None

            # Ideally we would also like to keep any formatting. This is easy if
            # there is only one paragraph. If there are multiple, we keep the
            paragraph_properties = None
            if is_paragraph:
                paragraph_properties = content.find('./w:p/w:pPr', XMLNS)

            # The properties of a run are not kept.
            # see ECMA-376 Part 1 § 17.5.2.27.
            #> If this element [std/stdPr/rpR] is not provided, the inserted is
            #> unformatted, as with any other run of text - it shall not inherit
            #> the properties of the placeholder text.
            run_properties = None

            # If the field is currently showing placeholder text, that text may
            # have different styling than is desired for the actual content, if
            # provided. (see ECMA-376 Part 1 § 17.5.2.27)
            placeholder_indicator = sdt.find('./w:sdtPr/w:showingPlcHdr', XMLNS)
            if placeholder_indicator is not None:
                # No longer displaying placeholder text, so remove the indicator.
                sdtPr = sdt.find('./w:sdtPr', XMLNS)
                assert sdtPr is not None # just fetched a sub-element of it
                sdtPr.remove(placeholder_indicator)

                # The element was showing placeholder text. Overwrite with
                # alternate properties/styling, if they are provided.
                run_properties = sdt.find('./w:sdtPr/w:rPr', XMLNS)

            # As mentioned above, there are many different possibilities here and
            # handling them all is not really needed for the use case here. So
            # instead we just remove everything and replace the sdrContent
            # child elements with something new.
            content.clear()

            # as mentioned earlier, we have to pay attention with paragraphs
            if is_paragraph:
                content = SubElementNS(content, 'w', 'p')
                # reinstate the paragraph properties/styling, if present
                if paragraph_properties is not None:
                    content.append(paragraph_properties)

            run = SubElementNS(content, 'w', 'r')
            # insert the run properties/styling, if given
            if run_properties is not None:
                run.append(run_properties)

            # now we finally get to adding the desired text content
            text = SubElementNS(run, 'w', 't')
            text.text = field_values[tag]

            # space=preserve is probably only rarely necessary, but adding it
            # also does not hurt, so we add it to avoid any weird cases where
            # it actually is necessary and expected to "just work".
            text.set('xml:space', 'preserve')

        # An important behaviour is that the output file is only created if the
        # process as a whole was successful. Thus we only copy the file now.

        # copy the prepared zip file to the output path
        shutil.copyfile(self._template_zip, output_path)
        # append the word/document.xml file to finalize
        with zipfile.ZipFile(output_path, 'a') as docx:
            with docx.open('word/document.xml', 'w') as document:
                # write the XML declaration ourself so we can properly get standalone=yes
                document.write(b'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\r\n')
                # since the XML declaration was already written, make sure it's not written again
                tree.write(document, encoding='UTF-8', xml_declaration=False)

if __name__ == '__main__':
	# This is an example how you can use the above code:
    TEMPLATE_PATH = 'template.docx'

    with MailMerge(TEMPLATE_PATH) as merge:
        merge({ "field": "example 1" }, 'output1.docx')
        merge({ "field": "example 2" }, 'output2.docx')
